# @package      hubzero-submit-distributor
# @file         JobStatus.py
# @author       Steven Clark <clarks@purdue.edu>
# @copyright    Copyright (c) 2012-2014 HUBzero Foundation, LLC.
# @license      http://www.gnu.org/licenses/lgpl-3.0.html LGPLv3
#
# Copyright (c) 2012-2014 HUBzero Foundation, LLC.
#
# This file is part of: The HUBzero(R) Platform for Scientific Collaboration
#
# The HUBzero(R) Platform for Scientific Collaboration (HUBzero) is free
# software: you can redistribute it and/or modify it under the terms of
# the GNU Lesser General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# HUBzero is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# HUBzero is a registered trademark of HUBzero Foundation, LLC.
#

__version__ = '2.7.1'

import sys
import os
import re
import signal
import subprocess
import logging
from errno import EPIPE

from hubzero.submit.CommandParser    import CommandParser
from hubzero.submit.RemoteJobMonitor import RemoteJobMonitor

class JobStatus:
   def __init__(self,
                jobMonitorHost,
                jobMonitorPort):

      self.logger           = logging.getLogger(__name__)
      self.version          = __version__
      self.operationMode    = 0
      self.jobMonitorHost   = jobMonitorHost
      self.jobMonitorPort   = jobMonitorPort
      self.remoteJobMonitor = RemoteJobMonitor(jobMonitorHost,jobMonitorPort)

      self.commandParser        = None
      self.doubleDashTerminator = bool(os.getenv("DOUBLE_DASH_TERMINATOR",'0'))
      self.statusIds            = []

      self.hubUserId = os.getuid()

      self.abortAttempted = False
      self.abortSignal    = 0

      signal.signal(signal.SIGINT,self.sigINT_handler)
      signal.signal(signal.SIGHUP,self.sigHUP_handler)
      signal.signal(signal.SIGQUIT,self.sigQUIT_handler)
      signal.signal(signal.SIGABRT,self.sigABRT_handler)
      signal.signal(signal.SIGTERM,self.sigTERM_handler)


   @staticmethod
   def __writeToStdout(message):
      try:
         sys.stdout.write(message)
         sys.stdout.flush()
      except IOError,err:
         pass


   @staticmethod
   def __writeToStderr(message):
      try:
         sys.stderr.write(message)
         sys.stderr.flush()
      except IOError,err:
         pass


   def parseCommandArguments(self):
      exitCode = 0
      self.commandParser = CommandParser(self.doubleDashTerminator)
      self.commandParser.parseArguments(sys.argv[1:])
      self.operationMode = self.commandParser.getOperationMode()

      if self.operationMode & self.commandParser.OPERATIONMODERUNSTATUS:
         self.statusIds = [str(statusId) for statusId in self.commandParser.getOption('statusIds')]
      else:
         exitCode = 1

      return(exitCode)

# SIGTERM is sent by Rappture Abort
# SIGHUP is sent by submit
# SIGHUP, SIGTERM are sent by session termination

   def sigGEN_handler(self,
                      signalNumber,
                      frame):
      if not self.abortAttempted:
         self.abortAttempted = True
         self.abortSignal    = signalNumber


   def sigINT_handler(self,
                      signalNumber,
                      frame):
      self.sigGEN_handler(signalNumber,frame)


   def sigHUP_handler(self,
                      signalNumber,
                      frame):
      self.sigGEN_handler(signalNumber,frame)


   def sigQUIT_handler(self,
                       signalNumber,
                       frame):
      self.sigGEN_handler(signalNumber,frame)


   def sigABRT_handler(self,
                       signalNumber,
                       frame):
      self.sigGEN_handler(signalNumber,frame)


   def sigTERM_handler(self,
                       signalNumber,
                       frame):
      self.sigGEN_handler(signalNumber,frame)


   @staticmethod
   def __natsort(l):
      def tryfloat(s):
         try:
            return int(s)
         except:
            return s.lower()

      def alphanum_key(s):
         """ Turn a string into a list of string and number chunks.
             "z23a" -> ["z", 23, "a"]
         """
         return [ tryfloat(c) for c in re.split('([0-9.]+)', s) ]

      """ Sort the given list in the way that humans expect.
      """
      l.sort(key=alphanum_key)


   def report(self):
      """Report job status"""

      reportTime,reportedJobs = self.remoteJobMonitor.queryUserActiveJobStatus(self.hubUserId)
      if len(reportedJobs) > 0:
         reportAllIds = len(self.statusIds) == 0
         maximumRunNameLength = len('RunName')
         localJobIds = reportedJobs.keys()
         for localJobId in localJobIds:
            if reportAllIds or (localJobId in self.statusIds):
               instanceIds = reportedJobs[localJobId].keys()
               for instanceId in instanceIds:
                  runName,jobQueue,jobSite,jobStatus,jobStage = reportedJobs[localJobId][instanceId]
                  maximumRunNameLength = max(maximumRunNameLength,len(runName))

         self.__writeToStdout("%-*s  %12s  %8s  %20s  %s\n" % (maximumRunNameLength,'RunName', \
                                                               'submitRun','Instance','Status','Location'))
         localJobIds = reportedJobs.keys()
         self.__natsort(localJobIds)
         for localJobId in localJobIds:
            if reportAllIds or (localJobId in self.statusIds):
               instanceIds = reportedJobs[localJobId].keys()
               self.__natsort(instanceIds)
               for instanceId in instanceIds:
                  runName,jobQueue,jobSite,jobStatus,jobStage = reportedJobs[localJobId][instanceId]
                  if jobQueue != '?':
                     queueSite = jobQueue + '@' + jobSite
                  else:
                     queueSite = jobSite
                  self.__writeToStdout("%-*s  %12s  %8s  %20s  %s\n" % (maximumRunNameLength,runName, \
                                                                        localJobId,instanceId,jobStatus,queueSite))

      del reportedJobs


